package footfall.streams;

import footfall.displays.*;
import java.applet.*;

/**
  Abstract class for generating stream of footfall via live input, file input, or simple modelling via FootModel generator.
  @author Chirag Mehta (chirag@chir.ag)
*/

public abstract class FootFallStream extends Applet implements Runnable
{
  // Display for footfall pressure data preview
  DisplayFootFall displayfootfall = null;

  // Thread info
  private volatile Thread thread;
  private volatile boolean threadDead = true;

  // Footfall data and parameters
  int padwidth = 16;
  int padheight = 44;
  float maxPressure = 0.0f;

  // Footfall animation and setup
  long startTime = 0;
  int framesPerSecond = 50;
  int framesPerSecondMin = 25;
  int framesPerSecondMax = 75;

  /**
    Abstract method to return the latest/current frame
  */
  public abstract float[][] getCurrentFrame();

  /**
    Frame height - default 44
  */
  public int getFrameHeight()
  {
    return this.padheight;
  }

  /**
    Frame width - default 16
  */
  public int getFrameWidth()
  {
    return this.padwidth;
  }

  /**
    Frames per second = Speed (25-75fps)
  */
  public int getFramesPerSecond()
  {
    return this.framesPerSecond;
  }

  /**
    Minimum frames per second - default 25
  */
  public int getFramesPerSecondMin()
  {
    return this.framesPerSecondMin;
  }

  /**
    Maximum frames per second - default 75
  */
  public int getFramesPerSecondMax()
  {
    return this.framesPerSecondMax;
  }

  /**
    The maximum pressure (0-50) for this stream. Weight in lbs is taken as roughly 10x the maximum pressure, hence 10lbs - 500lbs.
  */
  public float getMaxPressure()
  {
    return this.maxPressure;
  }

  /**
    Thread's run method to update the current frame depending on user-selected frames per second speed
  */
  public void run()
  {
    threadDead = false;
    Thread me = Thread.currentThread();

    while(thread == me)
    {
      update();
      try { thread.sleep(1000/getFramesPerSecond()); }
      catch(InterruptedException e) { }
    }

    thread = null;
    threadDead = true;
  }

  /**
    Start footfall stream generation
  */
  public void start()
  {
    thread = new Thread(this, "footfallstream");
    thread.setPriority(Thread.MAX_PRIORITY);
    thread.start();
  }

  /**
    Stop footfall stream generation
  */
  public void stop()
  {
    thread = null;
    while(!threadDead);
  }

  /**
    Set the display for preview of footfalls
    @param displayfootfall Footfall Display panel
  */
  public void setDisplay(DisplayFootFall displayfootfall)
  {
    this.displayfootfall = displayfootfall;
  }

  /**
    Set the speed - frames per second
    @param fps Frames per second
  */
  public void setFramesPerSecond(int fps)
  {
    this.framesPerSecond = fps;
  }

  /**
    Set the frame height for use in live recordings or modelling. Not used for recorded files.
    @param padheight Frame height
  */
  public void setFrameHeight(int padheight)
  {
    this.padheight = padheight;
  }

  /**
    Set the frame width for use in live recordings or modelling. Not used for recorded files.
    @param padwidth Frame width
  */
  public void setFrameWidth(int padwidth)
  {
    this.padwidth = padwidth;
  }

  /**
    Abstract method to set the maximum pressure (0-50) for this stream. Weight in lbs is taken as roughly 10x the maximum pressure, hence 10lbs - 500lbs.
    @param fMax Maximum pressure
  */
  public abstract void setMaxPressure(float fMax);

  /**
    Abstract method to generate the next frame based on footprints, speed, and weight (maximum pressure).
  */
  public abstract void update();
}