package footfall.streams;

import footfall.displays.*;
import java.applet.*;
import java.awt.*;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Vector;
import java.net.*;

/**
  Generate stream of footfall based on file input.
  @author Chirag Mehta (chirag@chir.ag)
*/

public class FootFallStreamFile extends FootFallStream
{
  // FootFall data and parameters
  private Vector frameNames, frameData;

  // Base and current frame numbers
  private int baseFrame = 0;
  private int currentFrame = 0;

  /**
    Initialize the footfall stream
  */
  public FootFallStreamFile()
  {
    this.frameNames = new Vector();
    this.frameData = new Vector();
  }

  /**
    Initialize the footfall stream and load a file
    @param filename Input file to load
  */
  public FootFallStreamFile(String filename)
  {
    this.frameNames = new Vector();
    this.frameData = new Vector();

    readFile(filename);
  }

  /**
    Returns the pressure data for the current frame
  */
  public float[][] getCurrentFrame()
  {
    return getFrame(getCurrentFrameNum());
  }

  /**
    Returns the current frame number
  */
  public int getCurrentFrameNum()
  {
    return this.currentFrame;
  }

  /**
    Returns the pressure data of the specific frame
    @param fNum Frame Number
  */
  public float[][] getFrame(int fNum)
  {
    return (float[][])this.frameData.elementAt(fNum);
  }

  /**
    Returns the name/title of the specific frame
    @param fNum Frame Number
  */
  public String getFrameName(int fNum)
  {
    return (String)this.frameNames.elementAt(fNum);
  }

  /**
    Returns number of frames in the current file
  */
  public int numFrames()
  {
    return this.frameNames.size();
  }

  /**
    Parses single line of ".bar" file format and loads it to buffer
    @param src Source string
  */
  private void parseBarLine(String src)
  {
    float[][] fL = new float[this.padheight][this.padwidth * 2];
    StringTokenizer row = new StringTokenizer(src, "|");

    for(int i=0; i < this.padheight && row.hasMoreTokens(); i++)
    {
      String rw = (String)row.nextToken();
      if(i == 0)
      {
        this.frameNames.add(rw);
      }
      else
      {
        StringTokenizer col = new StringTokenizer(rw, ",");

        for(int j = 0; j < padwidth *2 && col.hasMoreTokens(); j++)
        {
          fL[i-1][j] = Float.parseFloat(col.nextToken());
          if(fL[i-1][j] > this.maxPressure)
            this.maxPressure = fL[i-1][j];
        }
      }
    }

    this.frameData.add(fL);
  }

  /**
    Parses single frame of ".pad" or ".txt" file format and loads it to buffer
    @param fl[][] Frame data holder
    @param src Tab-delimited source frame
    @param i Frame number
    @param tab Left foot or right foot
  */
  private void parsePadLine(float fL[][], String src, int i, int tab)
  {
    StringTokenizer st = new StringTokenizer(src, "\t");

    for(int j = tab; j < tab + padwidth && st.hasMoreTokens(); j++)
    {
      fL[i][j] = Float.parseFloat(st.nextToken());
      if(fL[i][j] > this.maxPressure)
        this.maxPressure = fL[i][j];
    }
  }

  /**
    Prints the data values of an individual frame
    @param fNum Frame number to print
  */
  public void printFrame(int fNum)
  {
    float[][] frd = getFrame(fNum);

    System.out.println("\n\n" + getFrameName(fNum));

    System.out.println("\nFrame:\n");
    for(int i = 0; i < frd.length; i++)
    {
      for(int j = 0; j < frd[i].length; j++)
        System.out.print(frd[i][j] + ";");
      System.out.println("");
    }
  }

  /**
    Prints all the frames sequentially
  */
  public void printFrames()
  {
    for(int i = 0; i < numFrames(); i++)
      printFrame(i);
  }

  /**
    Detects file type and loads it
    @param filename Input file name in ".bar", ".pad", or ".txt" format
  */
  public void readFile(String filename)
  {
    if(filename.lastIndexOf(".bar") > -1)
      readBarFile(filename);
    else
      readPadFile(filename);
  }

  /**
    Reads the pressure data file and loads it into memory
    @param filename Input file name in ".bar" format
  */
  public void readBarFile(String filename)
  {
    BufferedReader input = null;

    this.frameNames.clear();
    this.frameData.clear();
    maxPressure = 0.0f;

    boolean appletMode = true;

    try
    {
      input = new BufferedReader(new FileReader(filename));
    }
    catch (Exception e)
    {
      try
      {
        ClassLoader cl = this.getClass().getClassLoader();
        URL url = cl.getResource(filename);
        InputStream is = url.openStream();
        input = new BufferedReader(new InputStreamReader(is));
      }
      catch (Exception ex) { }
    }

    try
    {
      String line;

      while ((line = input.readLine()) != null)
        parseBarLine(line);

      input.close();
    }
    catch (FileNotFoundException fnfe)
    {
      System.err.println("File not found: " + filename);
      System.exit(1);
    }
    catch (SecurityException se)
    {
      System.err.println("Cannot access file: " + filename);
      System.exit(1);
    }
    catch (IOException ioe)
    {
      System.err.println("Error reading file: " + filename);
      System.exit(1);
    }

    // Reset the start time of this file
    reset();
  }

  /**
    Reads the pressure data file and loads it into memory
    @param filename Input file name in ".pad" or ".txt" format
  */
  public void readPadFile(String filename)
  {
    BufferedReader input = null;

    this.frameNames.clear();
    this.frameData.clear();
    maxPressure = 0.0f;

    boolean appletMode = true;

    try
    {
      input = new BufferedReader(new FileReader(filename));
    }
    catch (Exception e)
    {
      try
      {
        ClassLoader cl = this.getClass().getClassLoader();
        URL url = cl.getResource(filename);
        InputStream is = url.openStream();
        input = new BufferedReader(new InputStreamReader(is));
      }
      catch (Exception ex) { }
    }

    try
    {
      String line = input.readLine();

      while (line != null)
      {
        String fName = line;
        line = input.readLine();

        float footData[][] = new float[padheight][padwidth * 2];
        for(int i = 0; i < padheight; i++)
        {
          line = input.readLine();
          parsePadLine(footData, line, i, 0);
        }
        line = input.readLine();

        for(int i = 0; i < padheight; i++)
        {
          line = input.readLine();
          parsePadLine(footData, line, i, padwidth);
        }
        line = input.readLine();
        line = input.readLine();

        this.frameNames.add(fName);
        this.frameData.add(footData);
      }

      input.close();

    }
    catch (FileNotFoundException fnfe)
    {
      System.err.println("File not found: " + filename);
      System.exit(1);
    }
    catch (SecurityException se)
    {
      System.err.println("Cannot access file: " + filename);
      System.exit(1);
    }
    catch (IOException ioe)
    {
      System.err.println("Error reading file: " + filename);
      System.exit(1);
    }

    // Reset the start time of this file
    reset();
  }

  /**
    Reads footprint frame from single string and creates a stream for modelling by FootModel
    @param line Foot print in ".bar" format
  */
  public void readPrint(String line)
  {
    this.frameNames.clear();
    this.frameData.clear();
    maxPressure = 0.0f;

    parseBarLine(line);
    reset();
  }

  /**
    Set current frame to beginning
  */
  public void reset()
  {
    setCurrentFrame(0);
  }

  /**
    Sets the base frame for rendering. Mostly used to set frame to beginning = 0.
    @param frm Frame number
  */
  public void setCurrentFrame(int frm)
  {
    this.baseFrame = frm;
    this.currentFrame = frm;
    this.startTime = System.currentTimeMillis();
  }

  /**
    Dummy function, does nothing; maximum pressure of recorded files is not changeable.
    @param fMax Ignored
  */
  public void setMaxPressure(float fMax)
  {
  }

  /**
    Set the current frame and draw the preview display
  */
  public void update()
  {
    long timeDif = Math.abs(System.currentTimeMillis() - this.startTime);
    int cf = (this.baseFrame + (int)(this.framesPerSecond * timeDif / 1000f)) % numFrames();

    this.currentFrame = cf;

    if(this.displayfootfall != null)
    {
      // Adjust for about 1kb*16/44100 second sound delay
      cf = (cf - (int)(this.framesPerSecond*16*1024/44100) + numFrames()) % numFrames();
      displayfootfall.updateDisplay(getFrame(cf), getMaxPressure(), getFrameWidth(), getFrameHeight(), getFrameName(cf));
    }
  }

  /**
    Saves the currently loaded frames as a ".bar" file format
    @param filename Name of the file to save to
  */
  public void writeBarFile(String filename)
  {
    BufferedWriter output = null;
    float[][] fr;
    String fName, fVal;

    try
    {
      output = new BufferedWriter(new FileWriter(filename));

      for(int k = 0; k < numFrames(); k++)
      {
        fr = getFrame(k);
        fName = getFrameName(k);
        output.write(fName + "|", 0, fName.length() + 1);
        for(int i = 0; i < getFrameHeight(); i++)
        {
          for(int j = 0; j < getFrameWidth() * 2; j++)
          {
            fVal = Float.toString(fr[i][j]);
            output.write(fVal, 0, fVal.length());
            if(j < getFrameWidth() * 2 - 1)
              output.write(",", 0, 1);
          }
          if(i < getFrameHeight() - 1)
            output.write("|", 0, 1);
        }
        output.newLine();
        output.flush();
      }

      output.close();

    }
    catch (SecurityException se)
    {
      System.err.println("Cannot access file: " + filename);
      System.exit(1);
    }
    catch (IOException ioe)
    {
      System.err.println("Error reading file: " + filename);
      System.exit(1);
    }
  }
}