package footfall.synths;

import footfall.generators.*;
import footfall.players.*;
import footfall.streams.*;
import java.util.Vector;

/**
  Synthesize footfall sounds based pressure data for each frame as a whole
  @author Chirag Mehta (chirag@chir.ag)
*/

public class FootFallSynthByFrame extends FootFallSynth
{
  /**
    Used for finding peaks and troughs in variance of consequent footfall frames. Looks back 5 frames. If peak is in 3rd frame, 2-frame delay of sound. Peaks/troughs are only used for wooden knocks and carpet thumps, not sand or snow.
  */
  protected StreamAnalyzer streamanalyzer = new StreamAnalyzer(5);

  // Frame data and variance for the preceding frame
  private float[][] lastFrame = null;
  private float lastVariance = 0;

  /**
    Synthesize footfall sounds based pressure data for each frame as a whole
  */
  public FootFallSynthByFrame()
  {
    super();
  }

  /**
    Synthesize footfall sounds based pressure data for each frame as a whole from specified FootFallStream
    @param footfallstream New footfall stream to get inputs from
  */
  public FootFallSynthByFrame(FootFallStream footfallstream)
  {
    super(footfallstream);
  }

  /**
    Perform granular synthesis for current frame as a whole.
  */
  public void synthesizeFootFallSound()
  {
    // Retrieve current frame for processing, exit if unavailable
    float currentFrame[][] = this.footfallstream.getCurrentFrame();
    if(currentFrame == null) return;

    // Maximum pressure value for current stream. Used for normalizing pressure values
    float maxPressure = this.footfallstream.getMaxPressure();

    // Bit rate for the player. Used for generating noise/waves at specific frequency
    int bitRate = grainplayer.getBitRate();

    // Length of grain - depends on bit rate and animation speed - frames per second
    int sampleDuration = bitRate/footfallstream.getFramesPerSecond();

    // Count of cells with non-zero pressure
    int nonBlankCells = 0;

    // Misc. variables used to calculating frequency and amplitude of grains
    float curAvg = 0, curMax = 0, curFactor = 0, curVariance = 0, curStdDev = 0;
    float difVariance = 0;
    float grainAmp = 0;
    int grainFreq = 0;

    // Find the current frame's average and maximum pressure values.
    // Also find the variance in pressure data among all cells.
    for(int i = 0; i < footfallstream.getFrameHeight(); i++)
      for(int j = 0; j < footfallstream.getFrameWidth() * 2; j++)
        if((curFactor = currentFrame[i][j]) > 0)
        {
          nonBlankCells++;
          curAvg += curFactor;
          curVariance += (curFactor * curFactor);
          if(curMax < curFactor) curMax = curFactor;
        }

    // If at least one cell has pressure > 0
    if(nonBlankCells > 1)
    {
      // Calculate variance, standard deviation, and average
      curVariance = ((nonBlankCells*curVariance) - (curAvg * curAvg))/(nonBlankCells*(nonBlankCells-1));
      curStdDev = (float)Math.sqrt(curVariance);
      curAvg /= (nonBlankCells > 0) ? nonBlankCells : 1;

      /**
        Sand
        - sampleDuration: Same for all grains - depends on bit rate and animation speed - frames per second
        - grainFreq: 44.1kHz * 0.4 = 17.6kHz
        - grainAmp : 0 - 0.1, depending on square of curFactor
        - curFactor: standard deviation is high for heels & toes, low when whole foot is down
      */
      curFactor = curMax/maxPressure * (1-Math.min(1/curStdDev, 1));
      grainFreq = (int)(bitRate*0.4f);
      grainAmp =  curFactor*curFactor*this.grainsize*0.1f;

      // Generate sound of sand when snow parameters are zero - white noise
      if(this.compression == 0 && this.crunchiness == 0 && grainAmp > 0)
        grainplayer.addGrain(soundgrainmodel.makeWhiteNoise(sampleDuration, grainFreq, grainAmp));

      /**
        If snow crunchiness or compression are turned on, generate lower frequency
        white noise sound for rudimentary simulation of splashing water.
        - curFactor: same as above
        - grainFreq: 44.1kHz * 0.02 = 882 Hz
        - grainAmp : same as above
      */
      grainFreq = (int)(bitRate*0.02f);
      if((this.compression > 0 || this.crunchiness > 0) && grainFreq > 0 && grainAmp > 0)
        grainplayer.addGrain(soundgrainmodel.makeWhiteNoise(sampleDuration, grainFreq, grainAmp));

      // Generate compression, crunching, and creakiness as difference from preceding frame
      if(lastFrame != null)
      {
        // Change in variance between current and preceding frame
        difVariance = curVariance-lastVariance;

        /**
          Compression (or compacting)
          - curFactor: high for few cells with high variance = heels/toes
                       low for low variance or high non-blank cells = flat foot
                       much more sensitive than curFactor above
          - k * 20   : 1kHz to 2kHz frequency at 44.1kHz bit rate
          - amplitude: depends on compression selection and curFactor-squared
        */
        curFactor = (float)Math.min(curVariance/nonBlankCells, 0.15f);
        if(difVariance > 0 && this.compression > 0 && curFactor > 0)
        {
          // Generate 50 grains at different frequencies - white noise
          for(int k = 50; k < 100; k++)
            grainplayer.addGrain(soundgrainmodel.makeWhiteNoise(
              sampleDuration,
              (int)(k*bitRate/2000),
              (float)(this.compression*curFactor*curFactor)
            ));
        }

        /**
          Crunching
          - curFactor: same as above
          - bitRate/k: frequency between 0.8kHz and 44.1kHz when played at 44.1kHz bit rate
          - amplitude: depends on crunchiness selection and curFactor-squared
        */
        if(difVariance != 0 && this.crunchiness > 0 && curFactor > 0)
        {
          // Generate 50 grains at different frequencies - white noise
          for(int k = 1; k < 50; k++)
            grainplayer.addGrain(soundgrainmodel.makeWhiteNoise(
              sampleDuration,
              (int)(bitRate/k),
              (float)(this.crunchiness*curFactor*curFactor)
            ));
        }

        /**
          Creaks - very short knocks or fading sine waves
          - curFactor: standard deviation is high for heels & toes, low when whole foot is down
          - bitRate/k: frequency between 4kHz and 9kHz when played at 44.1kHz bit rate
          - amplitude: depends on creakiness selection and curFactor
        */
        curFactor = curMax/maxPressure * (1-Math.min(1/curStdDev, 1));
        if(this.creakiness >= 0 && difVariance/lastVariance > 0.05)
          grainplayer.addGrain(soundgrainmodel.makeFadingSineWave(
            (int)(bitRate/400),
            (int)(((2-this.creakiness)*bitRate/100)+(curFactor*curFactor*bitRate/500)),
            (float)(this.creakiness*curFactor)
          ));
      }

      // If current variance is different from last variance
      if(streamanalyzer.add(curVariance))
      {
        // If a peak has been detected in the stream of variances
        if(streamanalyzer.isPeak())
        {
          streamanalyzer.clear();
          curFactor = (float)Math.sqrt(curMax/maxPressure);

          /**
            Carpet thumps - affected by speed
            - curFactor: find the frame's maximum pressure value and normalize it
            - frequency: 100-300hz
            - amplitude: depends on carpet selection and curFactor
          */
          if(this.carpet > 0 && curFactor > 0.85f)
            grainplayer.addGrain(soundgrainmodel.makeFadingSineWave(
              (int)(sampleDuration),
              (int)(50+this.carpet*50+(50*curFactor)),
              (float)(Math.sqrt(this.carpet)*curFactor)
            ));

          /**
            Knocks - constant length
            - curFactor: find the frame's maximum pressure value and normalize it
            - frequency: 150-750hz
            - amplitude: depends on woodenness selection and curFactor
          */
          if(this.woodenness > 0 && curFactor > 0.85f)
            grainplayer.addGrain(soundgrainmodel.makeFadingSineWave(
              (int)(bitRate/100),
              (int)(150+this.woodenness*250+(350*curFactor)),
              (float)(Math.sqrt(this.woodenness)*curFactor)
            ));
        }
      }
    }

    // Store the current frame and variance for next call
    lastFrame = currentFrame;
    lastVariance = curVariance;
  }

  /**
    Internal class used for finding peaks and troughs in variance of consequent footfall frames. Looks back 5 frames. If peak is in 3rd frame, 2-frame delay of sound. Peaks/troughs are only used for wooden knocks and carpet thumps. Sand and snow are 100% real-time.
  */
  private class StreamAnalyzer
  {
    // Vector to hold stream elements
    private Vector stream = new Vector();

    // Maximum length of the stream. Always odd number since middle element will be
    // checked for peak or trough.
    private int maxLength = 7;

    /**
      Start a new StreamAnalyzer with maxLength elements at most
      @param maxLength Maximum length of the stream. Must be an odd number.
    */
    public StreamAnalyzer(int maxLength)
    {
      this.maxLength = ((int)(maxLength/2))*2+1;
    }

    /**
      Add a new value to the stream and only keep the last maxLength values. Returns true when a unique element is added at the end.
      @param f New float value to append
    */
    public boolean add(float f)
    {
      if(stream.size() > 0 && ((Float)stream.lastElement()).floatValue() == f)
        return false;

      while(stream.size() >= maxLength)
        stream.remove(0);
      stream.add(new Float(f));

      return true;
    }

    /**
      Clears the stream for initialization after a peak or trough is found
    */
    public void clear()
    {
      stream.clear();
    }

    /**
      Returns the value of the middle element
    */
    public float getCenter()
    {
      if(stream.size() == 0)
        return 0;
      else if(stream.size() < (int)(maxLength/2+1))
        return ((Float)stream.lastElement()).floatValue();
      else
        return ((Float)stream.get((int)maxLength/2+1)).floatValue();
    }

    /**
      True if the middle element is a peak
    */
    public boolean isPeak()
    {
      if(stream.size() < maxLength)
        return false;

      float center = getCenter();

      for(int i = 0; i < (int)(maxLength/2); i++)
        if(((Float)stream.get(i)).floatValue() > center)
          return false;
      for(int i = (int)(maxLength/2)+1; i < maxLength; i++)
        if(((Float)stream.get(i)).floatValue() > center)
          return false;

      return true;
    }

    /**
      True if the middle element is a trough
    */
    public boolean isTrough()
    {
      if(stream.size() < maxLength)
        return false;

      float center = getCenter();

      for(int i = 0; i < (int)(maxLength/2); i++)
        if(((Float)stream.get(i)).floatValue() < center)
          return false;
      for(int i = (int)(maxLength/2)+1; i < maxLength; i++)
        if(((Float)stream.get(i)).floatValue() < center)
          return false;

      return true;
    }
  }
}